# Available platform types
AWS = "AWS"
AZURE = "AZURE"
GCP = "GCP"
AWS_GENERIC = "AWS_GENERIC"

# Available agent types
DATA_STORE_AGENT = "DATA_STORE_AGENT"
REMOTE_AGENT = "REMOTE_AGENT"

# Available authentication types
AZURE_STORAGE_ACCOUNT_KEYS = "AZURE_STORAGE_ACCOUNT_KEYS"
GCP_JSON_SERVICE_ACCOUNT_KEY = "GCP_JSON_SERVICE_ACCOUNT_KEY"
GCP_WORKLOAD_IDENTITY_FEDERATION = "GCP_WORKLOAD_IDENTITY_FEDERATION"
AWS_VPCE = "AWS_VPCE"
AWS_ASSUMABLE_ROLE = "AWS_ASSUMABLE_ROLE"
AZURE_FUNCTION_APP_KEY = "AZURE_FUNCTION_APP_KEY"
OAUTH2_CLIENT_CREDENTIALS = "OAUTH2_CLIENT_CREDENTIALS"
OAUTH2_PASSWORD = "OAUTH2_PASSWORD"

# Available storage types
AZURE_BLOB = "AZURE_BLOB"
GCS = "GCS"
S3 = "S3"

# Available wrapper types
TERRAFORM = "TERRAFORM"
