# Query responses
EXPECTED_GET_INSIGHTS_FIELD = "getInsights"
EXPECTED_GET_REPORT_FIELD = "getReportUrl"

# Insight Names
KEY_ASSETS_INSIGHT_NAME = "key_assets"
COVERAGE_OVERVIEW_INSIGHT_NAME = "monitors"
CLEANUP_SUGGESTIONS_INSIGHT_NAME = "cleanup_suggestions"
EVENTS_INSIGHT_NAME = "events"
READ_WRITE_INSIGHT_NAME = "table_read_write_stats"
INCIDENT_QUERY_INSIGHT_NAME = "incident_query_changes"
DETERIORATING_QUERY_INSIGHT_NAME = "query_runtime_trends"
RULES_AND_SLIS_INSIGHT_NAME = "rule_and_sli_results"

# Supported Schemes
FILE_SCHEME = "file"
S3_SCHEME = "s3"
SCHEME_DELIM = "://"

# Insights extension
INSIGHTS_DEFAULT_EXTENSION = ".csv"

# Insights headers
LIST_INSIGHTS_HEADERS = ["Insight (Name)", "Description", "Available"]
